__all__ = ["UniClient"]


from typing import overload

from unicex._abc import IUniClient
from unicex.enums import Exchange, OrderSide, OrderType, Timeframe
from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
)

from .adapter import Adapter
from .client import Client


class UniClient(IUniClient[Client]):
    """Унифицированный клиент для работы с Okx API."""

    @property
    def _client_cls(self) -> type[Client]:
        return Client

    async def tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.get_tickers(inst_type="SPOT")
        return Adapter.tickers(raw_data=raw_data, only_usdt=only_usdt)

    async def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.get_tickers(inst_type="SWAP")
        return Adapter.futures_tickers(raw_data=raw_data, only_usdt=only_usdt)

    async def last_price(self) -> dict[str, float]:
        raw_data = await self._client.get_tickers(inst_type="SPOT")
        return Adapter.last_price(raw_data)

    async def futures_last_price(self) -> dict[str, float]:
        raw_data = await self._client.get_tickers(inst_type="SWAP")
        return Adapter.last_price(raw_data)

    async def ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.get_tickers(inst_type="SPOT")
        return Adapter.ticker_24hr(raw_data=raw_data)

    async def futures_ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.get_tickers(inst_type="SWAP")
        return Adapter.futures_ticker_24hr(raw_data=raw_data)

    async def klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        interval = (
            interval.to_exchange_format(Exchange.OKX)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.get_candlesticks(
            inst_id=symbol,
            bar=interval,
            after=start_time,
            before=end_time,
            limit=limit,
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)

    async def futures_klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        interval = (
            interval.to_exchange_format(Exchange.OKX)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.get_candlesticks(
            inst_id=symbol,
            bar=interval,
            after=start_time,
            before=end_time,
            limit=limit,
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)

    @overload
    async def funding_rate(self, symbol: str) -> float: ...

    @overload
    async def funding_rate(self, symbol: None) -> dict[str, float]: ...

    @overload
    async def funding_rate(self) -> dict[str, float]: ...

    async def funding_rate(self, symbol: str = None) -> dict[str, float] | float:  # type: ignore[reportArgumentType]  # We want to raise our exception
        raw_data = await self._client.get_funding_rate("ANY")
        adapted_data = Adapter.funding_rate(raw_data)
        return adapted_data[symbol]

    @overload
    async def funding_interval(self, symbol: str) -> int: ...

    @overload
    async def funding_interval(self, symbol: None) -> dict[str, int]: ...

    @overload
    async def funding_interval(self) -> dict[str, int]: ...

    async def funding_interval(self, symbol: str | None = None) -> dict[str, int] | int:
        raw_data = await self._client.get_funding_rate(symbol if symbol else "ANY")
        adapted_data = Adapter.funding_interval(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    @overload
    async def open_interest(self, symbol: str) -> OpenInterestItem: ...

    @overload
    async def open_interest(self, symbol: None) -> OpenInterestDict: ...

    @overload
    async def open_interest(self) -> OpenInterestDict: ...

    async def open_interest(self, symbol: str | None = None) -> OpenInterestItem | OpenInterestDict:
        raw_data = await self._client.get_open_interest(inst_type="SWAP")
        adapted_data = Adapter.open_interest(raw_data)
        if symbol:
            return adapted_data[symbol]
        return adapted_data

    @overload
    async def futures_best_bid_ask(self, symbol: str) -> BestBidAskItem: ...

    @overload
    async def futures_best_bid_ask(self, symbol: None) -> BestBidAskDict: ...

    @overload
    async def futures_best_bid_ask(self) -> BestBidAskDict: ...

    async def futures_best_bid_ask(
        self, symbol: str | None = None
    ) -> BestBidAskItem | BestBidAskDict:
        raw_data = await self._client.get_tickers(inst_type="SWAP")
        adapted_data = Adapter.futures_best_bid_ask(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    async def futures_depth(
        self,
        symbol: str,
        limit: int,
    ) -> BookDepthDict:
        if not 1 <= limit <= 400:
            raise ValueError(f"Invalid limit for okx futures depth: {limit}. Valid range: [1, 400]")

        raw_data = await self._client.get_order_book(inst_id=symbol, sz=limit)
        return Adapter.futures_depth(raw_data=raw_data, symbol=symbol)

    async def futures_order_create(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        quantity: str,
        price: str | None = None,
        client_order_id: str | None = None,
        reduce_only: bool | None = None,
    ) -> OrderIdDict:
        raise NotImplementedError("Method will be implemented later.")

    async def futures_position_info(self, symbol: str) -> PositionInfoDict:
        raise NotImplementedError("Method will be implemented later.")



from __future__ import annotations

import time
from typing import Any

__all__ = ["Adapter"]

from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods

from .exchange_info import ExchangeInfo


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Gateio API."""

    @staticmethod
    def tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        return [
            item["currency_pair"]
            for item in raw_data
            if item["currency_pair"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def futures_tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        return [
            item["contract"]
            for item in raw_data
            if item["contract"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def last_price(raw_data: list[dict]) -> dict[str, float]:
        return {item["currency_pair"]: float(item["last"]) for item in raw_data}

    @staticmethod
    def futures_last_price(raw_data: list[dict]) -> dict[str, float]:
        return {item["contract"]: float(item["last"]) for item in raw_data}

    @staticmethod
    def ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        return {
            item["currency_pair"]: TickerDailyItem(
                p=float(item["change_percentage"]),
                v=float(item["base_volume"]),
                q=float(item["quote_volume"]),
            )
            for item in raw_data
        }

    @staticmethod
    def futures_ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        return {
            item["contract"]: TickerDailyItem(
                p=float(item["change_percentage"]),
                v=float(item["volume_24h_base"]),
                q=float(item["volume_24h_quote"]),
            )
            for item in raw_data
        }

    @staticmethod
    def klines(raw_data: list[list], symbol: str) -> list[KlineDict]:
        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]) * 1000,  # переводим секунды → миллисекунды
                o=float(kline[5]),
                h=float(kline[3]),
                l=float(kline[4]),
                c=float(kline[2]),
                v=float(kline[6]),
                q=float(kline[1]),
                T=None,
                x=kline[7] == "true",
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def futures_klines(raw_data: list[dict], symbol: str) -> list[KlineDict]:
        return [
            KlineDict(
                s=symbol,
                t=int(kline["t"]) * 1000,  # переводим секунды → миллисекунды
                o=float(kline["o"]),
                h=float(kline["h"]),
                l=float(kline["l"]),
                c=float(kline["c"]),
                v=float(kline["v"]),
                q=float(kline["sum"]),  # "sum" = объем в $ (quote volume)
                T=None,
                x=None,
            )
            for kline in sorted(raw_data, key=lambda x: int(x["t"]))
        ]

    @staticmethod
    def funding_rate(raw_data: list[dict]) -> dict[str, float]:
        return {
            item["contract"]: float(item["funding_rate"]) * 100
            for item in raw_data
            if item.get("funding_rate") is not None
        }

    @staticmethod
    def funding_interval(raw_data: list[dict]) -> dict[str, int]:
        return {
            item["name"]: int(item["funding_interval"]) // 3600
            for item in raw_data
            if item.get("funding_interval") is not None
        }

    @staticmethod
    def open_interest(raw_data: list[dict]) -> OpenInterestDict:
        return {
            item["contract"]: OpenInterestItem(
                t=int(time.time() * 1000),
                v=float(item["total_size"]) * float(item["quanto_multiplier"]),
                u="coins",
            )
            for item in raw_data
        }

    @staticmethod
    def futures_best_bid_ask(raw_data: list[dict]) -> BestBidAskDict:
        return {
            item["contract"]: BestBidAskItem(
                s=item["contract"],
                t=int(time.time() * 1000),  # REST endpoint не возвращает event time
                u=0,  # REST endpoint не возвращает update id
                b=float(item["highest_bid"]),
                B=float(item["highest_size"]) * Adapter._get_contract_size(item["contract"]),
                a=float(item["lowest_ask"]),
                A=float(item["lowest_size"]) * Adapter._get_contract_size(item["contract"]),
            )
            for item in raw_data
        }

    @staticmethod
    def futures_depth(raw_data: dict, symbol: str) -> BookDepthDict:
        contract_size = Adapter._get_contract_size(symbol)
        return BookDepthDict(
            s=symbol,
            t=int(float(raw_data["update"]) * 1000),
            u=int(raw_data.get("id", 0)),
            b=[(float(item["p"]), float(item["s"]) * contract_size) for item in raw_data["bids"]],
            a=[(float(item["p"]), float(item["s"]) * contract_size) for item in raw_data["asks"]],
        )

    @staticmethod
    def futures_order_create(raw_data: dict) -> OrderIdDict:
        return OrderIdDict(
            t=raw_data["create_time"] * 1000,
            id=raw_data["id"],
            cloid=raw_data["text"],
        )

    @staticmethod
    def futures_position_info(raw_data: dict) -> PositionInfoDict:
        if not raw_data:
            return PositionInfoDict(
                t=0,
                symbol="",
                side="",
                quantity=0,
                entry_price=0,
                mark_price=0,
                liquidation_price=0,
                unrealized_pnl=0,
                realized_pnl=0,
                leverage=0,
                notional=0,
            )
        contracts = float(raw_data["size"])
        contract_size = Adapter._get_contract_size(raw_data["contract"])
        quantity = abs(contracts) * contract_size

        return PositionInfoDict(
            t=raw_data["update_time"] * 1000,
            symbol=raw_data["contract"],
            side="BUY" if contracts > 0 else "SELL" if contracts < 0 else "",
            quantity=quantity,
            entry_price=float(raw_data["entry_price"]),
            mark_price=float(raw_data["mark_price"]),
            liquidation_price=float(raw_data["liq_price"]),
            unrealized_pnl=float(raw_data["unrealised_pnl"]),
            realized_pnl=float(raw_data["realised_pnl"]),
            leverage=float(raw_data["lever"]),
            notional=abs(float(raw_data["value"])),
        )

    @staticmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        data = raw_msg["result"]
        return [
            KlineDict(
                s=data["n"].split("_", 1)[1],  # XRP_USDT
                t=int(data["t"]) * 1000,
                o=float(data["o"]),
                h=float(data["h"]),
                l=float(data["l"]),
                c=float(data["c"]),
                v=float(data["a"]),
                q=float(data["v"]),
                T=None,
                x=not data["w"],  # w=False → свеча закрыта
            )
        ]

    @staticmethod
    def futures_klines_message(raw_msg: Any) -> list[KlineDict]:
        return [
            KlineDict(
                s=item["n"].split("_", 1)[1],  # XRP_USDT
                t=int(item["t"]) * 1000,
                o=float(item["o"]),
                h=float(item["h"]),
                l=float(item["l"]),
                c=float(item["c"]),
                v=float(item["a"]),
                q=float(item["v"]),
                T=None,
                x=not item["w"],  # w=False → свеча закрыта
            )
            for item in sorted(
                raw_msg["result"],
                key=lambda x: int(x["t"]),
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        trade = raw_msg["result"]
        return [
            TradeDict(
                t=trade["create_time_ms"],
                s=trade["currency_pair"],
                S=trade["side"].upper(),
                p=float(trade["price"]),
                v=float(trade["amount"]),
            )
        ]

    @staticmethod
    def futures_trades_message(raw_msg: Any) -> list[TradeDict]:
        return [
            TradeDict(
                t=item["create_time_ms"],
                s=item["contract"],
                S="BUY" if float(item["size"]) > 0 else "SELL",
                p=float(item["price"]),
                v=abs(float(item["size"])) * Adapter._get_contract_size(item["contract"]),
            )
            for item in sorted(
                raw_msg["result"],
                key=lambda x: x["create_time_ms"],
            )
        ]

    @staticmethod
    def futures_best_bid_ask_message(raw_msg: Any) -> list[BestBidAskItem]:
        result = raw_msg["result"]
        symbol = result["s"]
        contract_size = Adapter._get_contract_size(symbol)
        bid_price = float(result["b"]) if result["b"] != "" else 0.0
        ask_price = float(result["a"]) if result["a"] != "" else 0.0
        return [
            BestBidAskItem(
                s=symbol,
                t=int(result["t"]),
                u=int(result["u"]),
                b=bid_price,
                B=float(result["B"]) * contract_size,
                a=ask_price,
                A=float(result["A"]) * contract_size,
            )
        ]

    @staticmethod
    def futures_partial_book_depth_message(raw_msg: Any) -> list[BookDepthDict]:
        result = raw_msg["result"]
        symbol = result["contract"]
        contract_size = Adapter._get_contract_size(symbol)
        bids = result.get("bids", [])
        asks = result.get("asks", [])
        update_id = result.get("id", result.get("u"))
        if update_id is None:
            raise ValueError("Field `id` is missing in futures.order_book message")
        return [
            BookDepthDict(
                s=symbol,
                t=int(result["t"]),
                u=int(update_id),
                b=[(float(item["p"]), float(item["s"]) * contract_size) for item in bids],
                a=[(float(item["p"]), float(item["s"]) * contract_size) for item in asks],
            )
        ]

    @staticmethod
    def _get_contract_size(symbol: str) -> float:
        try:
            return ExchangeInfo.get_futures_ticker_info(symbol)["contract_size"] or 1
        except:  # noqa
            raise



__all__ = ["UniClient"]


from typing import overload

from unicex._abc import IUniClient
from unicex.enums import Exchange, OrderSide, OrderType, Timeframe
from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
)

from .adapter import Adapter
from .client import Client


class UniClient(IUniClient[Client]):
    """Унифицированный клиент для работы с Bybit API."""

    @property
    def _client_cls(self) -> type[Client]:
        return Client

    async def tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.tickers("spot")
        return Adapter.tickers(raw_data, only_usdt)

    async def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.tickers("linear")
        return Adapter.tickers(raw_data, only_usdt)

    async def last_price(self) -> dict[str, float]:
        raw_data = await self._client.tickers("spot")
        return Adapter.last_price(raw_data)

    async def futures_last_price(self) -> dict[str, float]:
        raw_data = await self._client.tickers("linear")
        return Adapter.last_price(raw_data)

    async def ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.tickers("spot")
        return Adapter.ticker_24hr(raw_data)

    async def futures_ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.tickers("linear")
        return Adapter.ticker_24hr(raw_data)

    async def klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        interval = (
            interval.to_exchange_format(Exchange.BYBIT)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.klines(
            category="spot",
            symbol=symbol,
            interval=interval,
            start=start_time,
            end=end_time,
            limit=limit,
        )
        return Adapter.klines(raw_data)

    async def futures_klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        interval = (
            interval.to_exchange_format(Exchange.BYBIT)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.klines(
            category="linear",
            symbol=symbol,
            interval=interval,
            start=start_time,
            end=end_time,
            limit=limit,
        )
        return Adapter.klines(raw_data)

    @overload
    async def funding_rate(self, symbol: str) -> float: ...

    @overload
    async def funding_rate(self, symbol: None) -> dict[str, float]: ...

    @overload
    async def funding_rate(self) -> dict[str, float]: ...

    async def funding_rate(self, symbol: str | None = None) -> dict[str, float] | float:
        raw_data = await self._client.tickers("linear", symbol=symbol)
        adapted_data = Adapter.funding_rate(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    @overload
    async def funding_interval(self, symbol: str) -> int: ...

    @overload
    async def funding_interval(self, symbol: None) -> dict[str, int]: ...

    @overload
    async def funding_interval(self) -> dict[str, int]: ...

    async def funding_interval(self, symbol: str | None = None) -> dict[str, int] | int:
        raw_data = await self._client.instruments_info(
            category="linear",
            symbol=symbol,
            limit=1000,
        )
        adapted_data = Adapter.funding_interval(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    @overload
    async def open_interest(self, symbol: str) -> OpenInterestItem: ...

    @overload
    async def open_interest(self, symbol: None) -> OpenInterestDict: ...

    @overload
    async def open_interest(self) -> OpenInterestDict: ...

    async def open_interest(self, symbol: str | None = None) -> OpenInterestItem | OpenInterestDict:
        raw_data = await self._client.tickers("linear")
        adapted_data = Adapter.open_interest(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    @overload
    async def futures_best_bid_ask(self, symbol: str) -> BestBidAskItem: ...

    @overload
    async def futures_best_bid_ask(self, symbol: None) -> BestBidAskDict: ...

    @overload
    async def futures_best_bid_ask(self) -> BestBidAskDict: ...

    async def futures_best_bid_ask(
        self, symbol: str | None = None
    ) -> BestBidAskItem | BestBidAskDict:
        raw_data = await self._client.tickers("linear", symbol=symbol)
        adapted_data = Adapter.futures_best_bid_ask(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    async def futures_depth(
        self,
        symbol: str,
        limit: int,
    ) -> BookDepthDict:
        if not 1 <= limit <= 500:
            raise ValueError(
                f"Invalid limit for bybit futures depth: {limit}. Valid range: [1, 500]"
            )

        raw_data = await self._client.orderbook(
            category="linear",
            symbol=symbol,
            limit=limit,
        )
        return Adapter.futures_depth(raw_data)

    async def futures_order_create(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        quantity: str,
        price: str | None = None,
        client_order_id: str | None = None,
        reduce_only: bool | None = None,
    ) -> OrderIdDict:
        self.ensure_authorized()

        raw_data = await self.client.create_order(
            category="linear",
            symbol=symbol,
            side=side.to_exchange_format(Exchange.BYBIT),  # type: ignore
            order_type=type.to_exchange_format(Exchange.BYBIT),  # type: ignore
            qty=quantity,
            price=price,
            order_link_id=client_order_id,
            reduce_only=reduce_only,
        )

        return Adapter.futures_order_create(raw_data)

    async def futures_position_info(self, symbol: str) -> PositionInfoDict:
        self.ensure_authorized()

        raw_data = await self._client.position_info(
            category="linear",
            symbol=symbol,
        )
        return Adapter.futures_position_info(raw_data)



__all__ = ["Adapter"]


from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    LiquidationDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Binance API."""

    @staticmethod
    def tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        return [
            item["symbol"] for item in raw_data if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        return {
            item["symbol"]: TickerDailyItem(
                p=float(item["priceChangePercent"]),
                q=float(item["quoteVolume"]),  # объём в долларах
                v=float(item["volume"]),  # объём в монетах
            )
            for item in raw_data
        }

    @staticmethod
    def last_price(raw_data: list[dict]) -> dict[str, float]:
        return {item["symbol"]: float(item["price"]) for item in raw_data}

    @staticmethod
    def klines(raw_data: list[list], symbol: str) -> list[KlineDict]:
        return [
            KlineDict(
                s=symbol,
                t=kline[0],
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[7]),
                T=kline[6],
                x=None,
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def funding_rate(raw_data: list[dict]) -> dict[str, float]:
        return {item["symbol"]: float(item["lastFundingRate"]) * 100 for item in raw_data}

    @staticmethod
    def funding_interval(raw_data: list[dict]) -> dict[str, int]:
        return {item["symbol"]: int(item["fundingIntervalHours"]) for item in raw_data}

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestItem:
        return OpenInterestItem(
            t=raw_data["time"],
            v=float(raw_data["openInterest"]),
            u="coins",
        )

    @staticmethod
    def futures_best_bid_ask(raw_data: list[dict]) -> BestBidAskDict:
        return {
            item["symbol"]: BestBidAskItem(
                s=item["symbol"],
                t=int(item["time"]),
                u=0,  # REST endpoint не возвращает update id
                b=float(item["bidPrice"]),
                B=float(item["bidQty"]),
                a=float(item["askPrice"]),
                A=float(item["askQty"]),
            )
            for item in raw_data
        }

    @staticmethod
    def futures_depth(raw_data: dict, symbol: str) -> BookDepthDict:
        return BookDepthDict(
            s=symbol,
            t=int(raw_data["E"]),
            u=int(raw_data["lastUpdateId"]),
            b=[(float(price), float(quantity)) for price, quantity in raw_data["bids"]],
            a=[(float(price), float(quantity)) for price, quantity in raw_data["asks"]],
        )

    @staticmethod
    def futures_order_create(raw_data: dict) -> OrderIdDict:
        return OrderIdDict(
            t=int(raw_data["updateTime"]),
            id=str(raw_data["orderId"]),
            cloid=str(raw_data.get("clientOrderId", "")),
        )

    @staticmethod
    def futures_position_info(raw_data: list[dict], symbol: str) -> PositionInfoDict:
        position = raw_data[0]
        position_amount = float(position["positionAmt"])
        return PositionInfoDict(
            t=int(position["updateTime"]),
            symbol=position["symbol"],
            side="BUY" if position_amount >= 0 else "SELL",
            quantity=abs(position_amount),
            entry_price=float(position["entryPrice"]),
            mark_price=float(position["markPrice"]),
            liquidation_price=float(position["liquidationPrice"]),
            unrealized_pnl=float(position["unRealizedProfit"]),
            realized_pnl=0,
            leverage=float(position["leverage"]),
            notional=abs(float(position["notional"])),
            breakeven_price=float(position.get("breakEvenPrice", 0)),
        )

    @staticmethod
    def klines_message(raw_msg: dict) -> list[KlineDict]:
        # Обрабатываем обертку в случае с multiplex stream
        kline = raw_msg.get("data", raw_msg)["k"]
        return [
            KlineDict(
                s=kline["s"],
                t=kline["t"],
                o=float(kline["o"]),
                h=float(kline["h"]),
                l=float(kline["l"]),
                c=float(kline["c"]),
                v=float(kline["v"]),  # Используем quote volume (в USDT)
                q=float(kline["q"]),  # Используем quote volume (в USDT)
                T=kline["T"],
                x=kline["x"],
            )
        ]

    @staticmethod
    def trades_message(raw_msg: dict) -> list[TradeDict]:
        msg = raw_msg.get("data", raw_msg)
        return [
            TradeDict(
                t=int(msg["T"]),
                s=str(msg["s"]),
                S="SELL" if bool(msg["m"]) else "BUY",
                p=float(msg["p"]),
                v=float(msg["q"]),
            )
        ]

    @staticmethod
    def liquidations_message(raw_msg: dict) -> list[LiquidationDict]:
        msg = raw_msg.get("data", raw_msg)
        liquidation = msg["o"]

        return [
            LiquidationDict(
                t=int(liquidation["T"]),
                s=str(liquidation["s"]),
                S="SHORT" if str(liquidation["S"]) == "BUY" else "LONG",
                v=float(liquidation["q"]),
                p=float(liquidation["ap"]),
            )
        ]

    @staticmethod
    def futures_best_bid_ask_message(raw_msg: dict) -> list[BestBidAskItem]:
        msg = raw_msg.get("data", raw_msg)

        return [
            BestBidAskItem(
                s=msg["s"],
                t=int(msg["E"]),
                u=int(msg["u"]),
                b=float(msg["b"]),
                B=float(msg["B"]),
                a=float(msg["a"]),
                A=float(msg["A"]),
            )
        ]

    @staticmethod
    def futures_partial_book_depth_message(raw_msg: dict) -> list[BookDepthDict]:
        msg = raw_msg.get("data", raw_msg)

        return [
            BookDepthDict(
                s=msg["s"],
                t=int(msg["E"]),
                u=int(msg["u"]),
                b=[(float(price), float(quantity)) for price, quantity in msg["b"]],
                a=[(float(price), float(quantity)) for price, quantity in msg["a"]],
            )
        ]



__all__ = ["Adapter"]

from collections.abc import Callable
from typing import Any

from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    LiquidationDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Bybit API."""

    @staticmethod
    def tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        return [
            item["symbol"]
            for item in raw_data["result"]["list"]
            if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: dict) -> TickerDailyDict:
        return {
            item["symbol"]: TickerDailyItem(
                p=round(float(item["price24hPcnt"]) * 100, 2),
                v=float(item["volume24h"]),
                q=float(item["turnover24h"]),
            )
            for item in raw_data["result"]["list"]
        }

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestDict:
        return {
            item["symbol"]: OpenInterestItem(
                t=raw_data["time"],
                v=float(item["openInterest"]),
                u="coins",
            )
            for item in raw_data["result"]["list"]
        }

    @staticmethod
    def funding_rate(raw_data: dict) -> dict[str, float]:
        return {
            item["symbol"]: float(item["fundingRate"]) * 100
            for item in raw_data["result"]["list"]
            if item["fundingRate"]
        }

    @staticmethod
    def funding_interval(raw_data: dict) -> dict[str, int]:
        return {
            item["symbol"]: int(item["fundingInterval"]) // 60
            for item in raw_data["result"]["list"]
        }

    @staticmethod
    def last_price(raw_data: dict) -> dict[str, float]:
        return {item["symbol"]: float(item["lastPrice"]) for item in raw_data["result"]["list"]}

    @staticmethod
    def futures_best_bid_ask(raw_data: dict) -> BestBidAskDict:
        return {
            item["symbol"]: BestBidAskItem(
                s=item["symbol"],
                t=int(raw_data["time"]),
                u=0,  # REST endpoint не возвращает update id
                b=float(item["bid1Price"]),
                B=float(item["bid1Size"]),
                a=float(item["ask1Price"]),
                A=float(item["ask1Size"]),
            )
            for item in raw_data["result"]["list"]
        }

    @staticmethod
    def futures_depth(raw_data: dict) -> BookDepthDict:
        result = raw_data["result"]
        return BookDepthDict(
            s=result["s"],
            t=int(result["ts"]),
            u=int(result["u"]),
            b=[(float(price), float(quantity)) for price, quantity in result["b"]],
            a=[(float(price), float(quantity)) for price, quantity in result["a"]],
        )

    @staticmethod
    def futures_order_create(raw_data: dict) -> OrderIdDict:
        result = raw_data["result"]
        return OrderIdDict(
            t=int(raw_data["time"]),
            id=result["orderId"],
            cloid=result.get("orderLinkId", "") or "",
        )

    @staticmethod
    def futures_position_info(raw_data: dict) -> PositionInfoDict:
        positions = raw_data["result"]["list"]
        if not positions:
            raise ValueError(f"Positions list are empty: {raw_data}.")

        position = positions[0]

        return PositionInfoDict(
            t=int(position.get("updatedTime") or raw_data["time"]),
            symbol=position["symbol"],
            side=position.get("side", "").upper(),
            quantity=float(position["size"]),
            entry_price=float(position.get("avgPrice") or 0.0),
            mark_price=float(position["markPrice"]),
            liquidation_price=float(position.get("liqPrice") or 0.0),
            unrealized_pnl=float(position.get("unrealisedPnl") or 0.0),
            realized_pnl=float(position.get("curRealisedPnl") or 0.0),
            leverage=float(position["leverage"]),
            notional=float(position.get("positionValue") or 0.0),
            breakeven_price=float(position.get("breakEvenPrice") or 0.0),
        )

    @staticmethod
    def klines(raw_data: dict) -> list[KlineDict]:
        return [
            KlineDict(
                s=raw_data["result"]["symbol"],
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[6]),
                T=None,
                x=None,
            )
            for kline in sorted(
                raw_data["result"]["list"],
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def Klines_message(raw_msg: Any) -> list[KlineDict]:
        symbol = raw_msg["topic"].split(".")[-1]
        return [
            KlineDict(
                s=symbol,
                t=kline["start"],
                o=float(kline["open"]),
                h=float(kline["high"]),
                l=float(kline["low"]),
                c=float(kline["close"]),
                v=float(kline["volume"]),
                q=float(kline["turnover"]),
                T=kline["end"],
                x=kline["confirm"],
            )
            for kline in sorted(
                raw_msg["data"],
                key=lambda x: int(x["start"]),
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        return [
            TradeDict(
                t=trade["T"],
                s=trade["s"],
                S=trade["S"].upper(),
                p=float(trade["p"]),
                v=float(trade["v"]),
            )
            for trade in sorted(
                raw_msg["data"],
                key=lambda x: int(x["T"]),
            )
        ]

    @staticmethod
    def liquidations_message(raw_msg: Any) -> list[LiquidationDict]:
        return [
            LiquidationDict(
                t=liquidation["T"],
                s=liquidation["s"],
                S="SHORT" if str(liquidation["S"]) == "buy" else "LONG",
                v=float(liquidation["v"]),
                p=float(liquidation["p"]),
            )
            for liquidation in sorted(
                raw_msg["data"],
                key=lambda x: int(x["T"]),
            )
        ]

    @staticmethod
    def futures_best_bid_ask_message(raw_msg: Any) -> list[BestBidAskItem]:
        data = raw_msg["data"]
        bid = data["b"][0]
        ask = data["a"][0]
        return [
            BestBidAskItem(
                s=data["s"],
                t=int(raw_msg["ts"]),
                u=int(data["u"]),
                b=float(bid[0]),
                B=float(bid[1]),
                a=float(ask[0]),
                A=float(ask[1]),
            )
        ]

    @staticmethod
    def futures_partial_book_depth_message() -> Callable[[Any], list[BookDepthDict]]:
        state: dict[str, dict[str, dict[float, float]]] = {}

        @catch_adapter_errors
        def _wrapper(raw_msg: Any) -> list[BookDepthDict]:
            data = raw_msg["data"]
            symbol = data["s"]

            symbol_state = state.setdefault(symbol, {"b": {}, "a": {}})
            bids_state = symbol_state["b"]
            asks_state = symbol_state["a"]

            if raw_msg.get("type") == "snapshot":
                bids_state.clear()
                asks_state.clear()

            for raw_price, raw_quantity in data["b"]:
                price = float(raw_price)
                quantity = float(raw_quantity)
                if quantity == 0.0:
                    bids_state.pop(price, None)
                else:
                    bids_state[price] = quantity

            for raw_price, raw_quantity in data["a"]:
                price = float(raw_price)
                quantity = float(raw_quantity)
                if quantity == 0.0:
                    asks_state.pop(price, None)
                else:
                    asks_state[price] = quantity

            return [
                BookDepthDict(
                    s=symbol,
                    t=int(raw_msg["ts"]),
                    u=int(data["u"]),
                    b=sorted(bids_state.items(), key=lambda item: item[0], reverse=True),
                    a=sorted(asks_state.items(), key=lambda item: item[0]),
                )
            ]

        return _wrapper



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo


class ExchangeInfo(IExchangeInfo):
    exchange_name = "BingX"
    """Название биржи, на которой работает класс."""

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None: ...

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None: ...



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo
from unicex.types import TickerInfoItem

from .client import Client


class ExchangeInfo(IExchangeInfo):
    exchange_name = "Hyperliquid"
    """Название биржи, на которой работает класс."""

    _spot_meta: dict = {}
    """Словарь с метаинформацией о спотовом рынке."""

    _spot_ident_to_idx: dict = {}
    """Словарь, в котором ключ - индетефикатор тикера на бирже, например '@123', а значение - его индекс в _spot_meta."""

    _spot_idx_to_name: dict = {}
    """Словарь, в котором ключ - индекс в _spot_meta, например "123", а значение - название тикера, например 'BTC'."""

    _futures_meta: dict = {}
    """Словарь с метаинформацией о фьючерсном рынке."""

    # DOCS: https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        cls._spot_meta = await Client(session).spot_metadata()
        cls._build_spot_mappings(cls._spot_meta)

        tickers_info: dict[str, TickerInfoItem] = {}
        for symbol_info in cls._spot_meta["tokens"]:
            try:
                tickers_info[symbol_info["name"]] = TickerInfoItem(
                    tick_step=None,
                    tick_precision=int(symbol_info["weiDecimals"]),
                    size_step=None,
                    size_precision=int(symbol_info["szDecimals"]),
                    contract_size=1,
                )
            except Exception as e:
                cls._logger.error(
                    f"{type(e)} creating TickerInfoItem for element={symbol_info}: {e}"
                )

        cls._tickers_info = tickers_info

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        cls._futures_meta = await Client(session).perp_metadata()

    @classmethod
    def _build_spot_mappings(cls, spot_meta: dict) -> None:
        universe = spot_meta["universe"]
        tokens = spot_meta["tokens"]

        number_to_idx = {}
        for u in universe:
            number_to_idx[u["name"]] = u["tokens"][0]
        cls._spot_ident_to_idx = number_to_idx

        idx_to_name = {}
        for t in tokens:
            idx_to_name[t["index"]] = t["name"]
        cls._spot_idx_to_name = idx_to_name

    @classmethod
    def get_spot_meta(cls) -> dict:
        cls._check_loaded()
        return cls._spot_meta

    @classmethod
    def get_futures_meta(cls) -> dict:
        cls._check_loaded()
        return cls._futures_meta

    @classmethod
    def resolve_spot_symbol(cls, ident: str) -> str:
        cls._check_loaded()

        try:
            return cls._spot_idx_to_name[cls._spot_ident_to_idx[ident]]
        except KeyError:
            return ident

    @classmethod
    def resolve_spot_ident(cls, symbol: str) -> str:
        cls._check_loaded()

        # Находим индекс по тикеру
        idx = next(k for k, v in cls._spot_idx_to_name.items() if v == symbol)

        # Возвращаем внутренний идентификатор вида "@142"
        return next(k for k, v in cls._spot_ident_to_idx.items() if v == idx)



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo
from unicex.types import TickerInfoItem

from .client import Client


class ExchangeInfo(IExchangeInfo):
    exchange_name = "Mexc"
    """Название биржи, на которой работает класс."""

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        exchange_info = await Client(session).exchange_info()
        tickers_info = {}
        for el in exchange_info["symbols"]:
            try:
                tickers_info[el["symbol"]] = TickerInfoItem(
                    tick_precision=int(el["quotePrecision"]),
                    tick_step=None,
                    size_precision=int(el["baseAssetPrecision"]),
                    size_step=None,
                    contract_size=1,
                )
            except Exception as e:
                cls._logger.error(f"{type(e)} creating TickerInfoItem for element={el}: {e}")

        cls._tickers_info = tickers_info

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        exchange_info = await Client(session).futures_contract_detail()
        tickers_info = {}
        for el in exchange_info["data"]:
            try:
                tickers_info[el["symbol"]] = TickerInfoItem(
                    tick_precision=None,
                    tick_step=el["priceUnit"],
                    size_precision=None,
                    size_step=el["contractSize"],
                    contract_size=el["contractSize"],
                )
            except Exception as e:
                cls._logger.error(f"{type(e)} creating TickerInfoItem for element={el}: {e}")

        cls._futures_tickers_info = tickers_info



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo
from unicex.types import TickerInfoItem

from .client import Client


class ExchangeInfo(IExchangeInfo):
    exchange_name = "Gateio"
    """Название биржи, на которой работает класс."""

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        # У Gate список пар очень большой, поэтому даем больше времени на ответ и чтение.
        client = Client(session, timeout=60, max_retries=5, retry_delay=0.5)
        currency_pairs = await client.currency_pairs()
        tickers_info: dict[str, TickerInfoItem] = {}
        for symbol_info in currency_pairs:
            try:
                tickers_info[symbol_info.get("id")] = TickerInfoItem(
                    tick_precision=int(symbol_info["precision"]),
                    tick_step=None,
                    size_precision=int(symbol_info["amount_precision"]),
                    size_step=None,
                    contract_size=1,
                )
            except Exception as e:
                cls._logger.error(
                    f"{type(e)} creating TickerInfoItem for element={symbol_info}: {e}"
                )

        cls._tickers_info = tickers_info

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        client = Client(session, timeout=60, max_retries=5, retry_delay=0.5)
        contracts = await client.futures_contracts("usdt")
        tickers_info: dict[str, TickerInfoItem] = {}
        for contract in contracts:
            try:
                tickers_info[contract.get("name")] = TickerInfoItem(
                    tick_precision=None,
                    tick_step=float(contract["order_price_round"]),
                    size_precision=None,
                    size_step=float(contract["quanto_multiplier"]),
                    contract_size=float(contract["quanto_multiplier"]),
                )
            except Exception as e:
                cls._logger.error(f"{type(e)} creating TickerInfoItem for element={contract}: {e}")

        cls._futures_tickers_info = tickers_info



__all__ = ["Adapter"]


from typing import Any

from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods

from .exchange_info import ExchangeInfo


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Okx API."""

    @staticmethod
    def tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        return [
            item["instId"]
            for item in raw_data["data"]
            if item["instId"].endswith("-USDT") or not only_usdt
        ]

    @staticmethod
    def futures_tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        return [
            item["instId"]
            for item in raw_data["data"]
            if item["instId"].endswith("-USDT-SWAP") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: dict) -> TickerDailyDict:
        # Обратите внимание, изменение цены в случае с OKX возвращается относительно открытия 1 day свечи.
        result = {}
        for item in raw_data["data"]:
            try:
                result[item["instId"]] = TickerDailyItem(
                    p=round(
                        (float(item["last"]) - float(item["open24h"]))
                        / float(item["open24h"])
                        * 100,
                        2,
                    ),
                    v=float(item["vol24h"]),
                    q=float(item["volCcy24h"]),
                )
            except (ValueError, TypeError, KeyError):
                continue
        return result

    @staticmethod
    def futures_ticker_24hr(raw_data: dict) -> TickerDailyDict:
        # Обратите внимание, изменение цены в случае с OKX возвращается относительно открытия 1 day свечи.
        result = {}
        for item in raw_data["data"]:
            try:
                last = float(item["last"])
                open_24h = float(item["open24h"])
                vol_ccy_24h = float(item["volCcy24h"])
                result[item["instId"]] = TickerDailyItem(
                    p=round((last - open_24h) / open_24h * 100, 2),
                    v=vol_ccy_24h,
                    q=vol_ccy_24h * last,
                )
            except (ValueError, TypeError, KeyError, ZeroDivisionError):
                continue
        return result

    @staticmethod
    def last_price(raw_data: dict) -> dict[str, float]:
        result = {}
        for item in raw_data["data"]:
            try:
                result[item["instId"]] = float(item["last"])
            except (ValueError, TypeError, KeyError):
                continue
        return result

    @staticmethod
    def klines(raw_data: dict, symbol: str) -> list[KlineDict]:
        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[6]),
                q=float(kline[7]),
                T=None,
                x=bool(int(kline[8])),
            )
            for kline in sorted(
                raw_data["data"],
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def funding_rate(raw_data: dict) -> dict[str, float]:
        data = raw_data["data"][0]
        return {data["instId"]: float(data["fundingRate"]) * 100}

    @staticmethod
    def funding_interval(raw_data: dict) -> dict[str, int]:
        return {
            item["instId"]: (int(item["fundingTime"]) - int(item["prevFundingTime"]))
            // (1000 * 60 * 60)
            for item in raw_data["data"]
            if item.get("fundingTime") and item.get("prevFundingTime")
        }

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestDict:
        return {
            item["instId"]: OpenInterestItem(
                t=int(item["ts"]),
                v=float(item["oiCcy"]),
                u="coins",
            )
            for item in raw_data["data"]
        }

    @staticmethod
    def futures_best_bid_ask(raw_data: dict) -> BestBidAskDict:
        return {
            item["instId"]: BestBidAskItem(
                s=item["instId"],
                t=int(item["ts"]),
                u=0,  # REST endpoint не возвращает update id
                b=float(item["bidPx"]),
                B=float(item["bidSz"]) * Adapter._get_contract_size(item["instId"]),
                a=float(item["askPx"]),
                A=float(item["askSz"]) * Adapter._get_contract_size(item["instId"]),
            )
            for item in raw_data["data"]
        }

    @staticmethod
    def futures_depth(raw_data: dict, symbol: str) -> BookDepthDict:
        data = raw_data["data"][0]
        contract_size = Adapter._get_contract_size(symbol)
        return BookDepthDict(
            s=symbol,
            t=int(data["ts"]),
            u=int(data["seqId"]),
            b=[(float(item[0]), float(item[1]) * contract_size) for item in data["bids"]],
            a=[(float(item[0]), float(item[1]) * contract_size) for item in data["asks"]],
        )

    @staticmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        return [
            KlineDict(
                s=raw_msg["arg"]["instId"],
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[6]),
                q=float(kline[7]),
                T=None,
                x=bool(int(kline[8])),
            )
            for kline in sorted(raw_msg["data"], key=lambda item: int(item[0]))
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        return [
            TradeDict(
                t=int(trade["ts"]),
                s=trade["instId"],
                S=trade["side"].upper(),
                p=float(trade["px"]),
                v=float(trade["sz"]) * Adapter._get_contract_size(trade["instId"]),
            )
            for trade in sorted(raw_msg["data"], key=lambda item: int(item["ts"]))
        ]

    @staticmethod
    def futures_best_bid_ask_message(raw_msg: Any) -> list[BestBidAskItem]:
        data = raw_msg["data"][0]
        inst_id = raw_msg["arg"]["instId"]
        contract_size = Adapter._get_contract_size(inst_id)
        bids = data.get("bids", [])
        asks = data.get("asks", [])

        best_bid = bids[0] if bids else None
        best_ask = asks[0] if asks else None

        bid_price = float(best_bid[0]) if best_bid else 0.0
        bid_size = float(best_bid[1]) if best_bid else 0.0
        ask_price = float(best_ask[0]) if best_ask else 0.0
        ask_size = float(best_ask[1]) if best_ask else 0.0

        return [
            BestBidAskItem(
                s=str(raw_msg["arg"]["instId"]),
                t=int(data["ts"]),
                u=int(data["seqId"]),
                b=bid_price,
                B=bid_size * contract_size,
                a=ask_price,
                A=ask_size * contract_size,
            )
        ]

    @staticmethod
    def futures_partial_book_depth_message(raw_msg: Any) -> list[BookDepthDict]:
        data = raw_msg["data"][0]
        inst_id = raw_msg["arg"]["instId"]
        contract_size = Adapter._get_contract_size(inst_id)
        bids = data.get("bids", [])
        asks = data.get("asks", [])

        return [
            BookDepthDict(
                s=inst_id,
                t=int(data["ts"]),
                u=int(data["seqId"]),
                b=[(float(item[0]), float(item[1]) * contract_size) for item in bids],
                a=[(float(item[0]), float(item[1]) * contract_size) for item in asks],
            )
        ]

    @staticmethod
    def _get_contract_size(symbol: str) -> float:
        try:
            return ExchangeInfo.get_futures_ticker_info(symbol)["contract_size"] or 1
        except:  # noqa
            return 1



__all__ = ["IUniWebsocketManager"]

from collections.abc import Awaitable, Callable, Sequence
from typing import Any

from unicex._abc import IUniWebsocketManager
from unicex._base import Websocket
from unicex.enums import Exchange, MarketType, Timeframe
from unicex.types import LoggerLike

from .adapter import Adapter
from .client import Client
from .uni_client import UniClient
from .websocket_manager import WebsocketManager

type CallbackType = Callable[[Any], Awaitable[None]]


class UniWebsocketManager(IUniWebsocketManager):
    """Реализация менеджера асинхронных унифицированных вебсокетов."""

    def __init__(
        self,
        client: Client | UniClient | None = None,
        logger: LoggerLike | None = None,
        **ws_kwargs: Any,
    ) -> None:
        """Инициализирует унифицированный менеджер вебсокетов.

        Параметры:
            client (`Client | UniClient | None`): Клиент Gateio или унифицированный клиент. Нужен для подключения к приватным топикам.
            logger (`LoggerLike | None`): Логгер для записи логов.
            ws_kwargs (`dict[str, Any]`): Дополнительные параметры инициализации, которые будут переданы WebsocketManager/Websocket.
        """
        super().__init__(client=client, logger=logger)
        self._websocket_manager = WebsocketManager(self._client, **ws_kwargs)  # type: ignore
        self._adapter = Adapter()

    def _is_service_message(self, raw_msg: Any) -> bool:
        is_sub_msg = raw_msg.get("event") == "subscribe"
        is_pong_msg = raw_msg.get("event") == "pong"
        is_pong_msg_2 = raw_msg.get("channel") in ["spot.pong", "futures.pong"]
        return is_sub_msg or is_pong_msg or is_pong_msg_2

    def _normalize_symbols(
        self,
        symbol: str | None,
        symbols: Sequence[str] | None,
    ) -> list[str]:
        if symbol and symbols:
            raise ValueError("Parameters symbol and symbols cannot be used together")
        if symbol:
            return [symbol]
        if symbols:
            return list(symbols)
        raise ValueError("Either symbol or symbols must be provided")

    def klines(
        self,
        callback: CallbackType,
        timeframe: Timeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        tickers = self._normalize_symbols(symbol, symbols)

        wrapper = self._make_wrapper(self._adapter.klines_message, callback)
        return self._websocket_manager.candlesticks(
            callback=wrapper,
            interval=timeframe.to_exchange_format(Exchange.GATE, MarketType.SPOT),
            symbols=tickers,
        )

    def futures_klines(
        self,
        callback: CallbackType,
        timeframe: Timeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        tickers = self._normalize_symbols(symbol, symbols)

        wrapper = self._make_wrapper(self._adapter.futures_klines_message, callback)
        return self._websocket_manager.futures_candlesticks(
            callback=wrapper,
            interval=timeframe.to_exchange_format(Exchange.GATE, MarketType.FUTURES),
            symbols=tickers,
        )

    def trades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        tickers = self._normalize_symbols(symbol, symbols)

        wrapper = self._make_wrapper(self._adapter.trades_message, callback)
        return self._websocket_manager.trades(callback=wrapper, symbols=tickers)

    def aggtrades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        return self.trades(callback=callback, symbol=symbol, symbols=symbols)  # type: ignore[reportCallIssue]

    def futures_trades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        tickers = self._normalize_symbols(symbol, symbols)

        wrapper = self._make_wrapper(self._adapter.futures_trades_message, callback)
        return self._websocket_manager.futures_trades(callback=wrapper, symbols=tickers)

    def futures_aggtrades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        return self.futures_trades(callback=callback, symbol=symbol, symbols=symbols)  # type: ignore[reportCallIssue]

    def futures_best_bid_ask(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        tickers = self._normalize_symbols(symbol, symbols)

        wrapper = self._make_wrapper(self._adapter.futures_best_bid_ask_message, callback)
        return self._websocket_manager.futures_book_ticker(callback=wrapper, symbols=tickers)

    def futures_partial_book_depth(
        self,
        callback: CallbackType,
        limit: int,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        allowed_levels = {1, 5, 10, 20, 50, 100}
        if limit not in allowed_levels:
            raise ValueError("Parameter `limit` must be one of: 1, 5, 10, 20, 50, 100")

        tickers = self._normalize_symbols(symbol, symbols)

        wrapper = self._make_wrapper(self._adapter.futures_partial_book_depth_message, callback)
        return self._websocket_manager.futures_order_book(
            callback=wrapper,
            symbols=tickers,
            limit=str(limit),  # type: ignore[arg-type]
            interval="0",
        )



from typing import Any

from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Bitget API."""

    @staticmethod
    def tickers(raw_data: Any, only_usdt: bool) -> list[str]:
        return [
            item["symbol"]
            for item in raw_data["data"]
            if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: Any) -> TickerDailyDict:
        return {
            item["symbol"]: TickerDailyItem(
                p=round(float(item["change24h"]) * 100, 2),  # конвертируем в проценты
                v=float(item["baseVolume"]),  # объём в COIN
                q=float(item["usdtVolume"]),  # объём в USDT
            )
            for item in raw_data["data"]
        }

    @staticmethod
    def last_price(raw_data: Any) -> dict[str, float]:
        return {item["symbol"]: float(item["lastPr"]) for item in raw_data["data"]}

    @staticmethod
    def klines(raw_data: Any, symbol: str) -> list[KlineDict]:
        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[6]),
                T=None,
                x=None,
            )
            for kline in sorted(
                raw_data["data"],
                key=lambda x: int(x[0]),  # Bitget присылает пачку трейдов в обратном порядке
            )
        ]

    @staticmethod
    def funding_rate(raw_data: Any) -> dict[str, float]:
        return {item["symbol"]: float(item["fundingRate"]) * 100 for item in raw_data["data"]}

    @staticmethod
    def funding_interval(raw_data: Any) -> dict[str, int]:
        return {item["symbol"]: int(item["fundingRateInterval"]) for item in raw_data["data"]}

    @staticmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        symbol = raw_msg["arg"]["instId"]

        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[6]),
                T=None,
                x=None,
            )
            for kline in sorted(
                raw_msg["data"],
                key=lambda x: int(x[0]),  # Bitget присылает пачку трейдов в обратном порядке
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        symbol = raw_msg["arg"]["instId"]

        return [
            TradeDict(
                t=int(trade["ts"]),
                s=symbol,
                S=trade["side"].upper(),
                p=float(trade["price"]),
                v=float(trade["size"]),
            )
            for trade in sorted(
                raw_msg["data"],
                key=lambda x: int(x["ts"]),  # Bitget присылает пачку трейдов в обратном порядке
            )
        ]

    @staticmethod
    def open_interest(raw_data: Any) -> OpenInterestDict:
        return {
            i["symbol"]: OpenInterestItem(
                t=int(i["ts"]),
                v=float(i["holdingAmount"]),
                u="coins",
            )
            for i in raw_data["data"]
        }

    @staticmethod
    def futures_best_bid_ask(raw_data: Any) -> BestBidAskDict:
        return {
            item["symbol"]: BestBidAskItem(
                s=item["symbol"],
                t=int(item["ts"]),
                u=0,  # REST endpoint не возвращает update id
                b=float(item["bidPr"]),
                B=float(item["bidSz"]),
                a=float(item["askPr"]),
                A=float(item["askSz"]),
            )
            for item in raw_data["data"]
        }

    @staticmethod
    def futures_depth(raw_data: Any, symbol: str) -> BookDepthDict:
        data = raw_data["data"]
        return BookDepthDict(
            s=symbol,
            t=int(data["ts"]),
            u=0,  # REST endpoint не возвращает update id
            b=[(float(price), float(quantity)) for price, quantity in data["bids"]],
            a=[(float(price), float(quantity)) for price, quantity in data["asks"]],
        )

    @staticmethod
    def futures_order_create(raw_data: Any) -> OrderIdDict:
        data = raw_data["data"]
        return OrderIdDict(
            t=int(raw_data["requestTime"]),
            id=data["orderId"],
            cloid=data["clientOid"],
        )

    @staticmethod
    def futures_position_info(raw_data: Any) -> PositionInfoDict:
        positions = raw_data["data"]
        if not positions:
            return PositionInfoDict(
                t=raw_data["requestTime"],
                symbol="",
                side="",
                quantity=0,
                entry_price=0,
                mark_price=0,
                liquidation_price=0,
                unrealized_pnl=0,
                realized_pnl=0,
                leverage=0,
            )

        position = positions[0]

        return PositionInfoDict(
            t=int(position["uTime"]),
            symbol=position["symbol"],
            side={"long": "BUY", "short": "SELL"}[position["holdSide"]],  # type: ignore
            quantity=float(position["total"]),
            entry_price=float(position["openPriceAvg"]),
            mark_price=float(position["markPrice"]),
            liquidation_price=float(position["liquidationPrice"]),
            unrealized_pnl=float(position["unrealizedPL"]),
            realized_pnl=float(position["achievedProfits"]) - float(position["deductedFee"]),
            leverage=float(position["leverage"]),
            breakeven_price=float(position["breakEvenPrice"]),
        )

    @staticmethod
    def futures_best_bid_ask_message(raw_msg: Any) -> list[BestBidAskItem]:
        arg = raw_msg["arg"]
        data = raw_msg["data"][0]
        best_bid = data["bids"][0]
        best_ask = data["asks"][0]
        return [
            BestBidAskItem(
                s=arg["instId"],
                t=int(data["ts"]),
                u=int(data["seq"]),
                b=float(best_bid[0]),
                B=float(best_bid[1]),
                a=float(best_ask[0]),
                A=float(best_ask[1]),
            )
        ]

    @staticmethod
    def futures_partial_book_depth_message(raw_msg: Any) -> list[BookDepthDict]:
        data = raw_msg["data"][0]
        return [
            BookDepthDict(
                s=raw_msg["arg"]["instId"],
                t=int(data["ts"]),
                u=int(data["seq"]),
                b=[(float(price), float(quantity)) for price, quantity in data["bids"]],
                a=[(float(price), float(quantity)) for price, quantity in data["asks"]],
            )
        ]



__all__ = ["UniClient"]


from decimal import Decimal
from typing import Literal, overload

from unicex._abc import IUniClient
from unicex.enums import Exchange, MarketType, OrderSide, OrderType, Timeframe
from unicex.exceptions import ResponseError
from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
)

from .adapter import Adapter
from .client import Client
from .exchange_info import ExchangeInfo


class UniClient(IUniClient[Client]):
    """Унифицированный клиент для работы с Gateio API."""

    @property
    def _client_cls(self) -> type[Client]:
        return Client

    async def tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.tickers()
        return Adapter.tickers(raw_data=raw_data, only_usdt=only_usdt)  # type: ignore[reportArgumentType]

    async def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.futures_tickers(settle="usdt")
        return Adapter.futures_tickers(raw_data=raw_data, only_usdt=only_usdt)  # type: ignore[reportArgumentType]

    async def last_price(self) -> dict[str, float]:
        raw_data = await self._client.tickers()
        return Adapter.last_price(raw_data=raw_data)  # type: ignore[reportArgumentType]

    async def futures_last_price(self) -> dict[str, float]:
        raw_data = await self._client.futures_tickers(settle="usdt")
        return Adapter.futures_last_price(raw_data=raw_data)  # type: ignore[reportArgumentType]

    async def ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.tickers()
        return Adapter.ticker_24hr(raw_data=raw_data)  # type: ignore[reportArgumentType]

    async def futures_ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.futures_tickers(settle="usdt")
        return Adapter.futures_ticker_24hr(raw_data=raw_data)  # type: ignore[reportArgumentType]

    async def klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        interval = (
            interval.to_exchange_format(Exchange.GATE, MarketType.SPOT)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.candlesticks(
            currency_pair=symbol,
            interval=interval,
            limit=limit,
            from_time=self._to_seconds(start_time),
            to_time=self._to_seconds(end_time),
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)  # type: ignore[reportArgumentType]

    async def futures_klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        interval = (
            interval.to_exchange_format(Exchange.GATE, MarketType.FUTURES)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.futures_candlesticks(
            settle="usdt",
            contract=symbol,
            interval=interval,
            limit=limit,
            from_time=self._to_seconds(start_time),
            to_time=self._to_seconds(end_time),
        )
        return Adapter.futures_klines(raw_data=raw_data, symbol=symbol)  # type: ignore[reportArgumentType]

    @overload
    async def funding_rate(self, symbol: str) -> float: ...

    @overload
    async def funding_rate(self, symbol: None) -> dict[str, float]: ...

    @overload
    async def funding_rate(self) -> dict[str, float]: ...

    async def funding_rate(self, symbol: str | None = None) -> dict[str, float] | float:
        raw_data = await self._client.futures_tickers(settle="usdt", contract=symbol)
        items = raw_data if isinstance(raw_data, list) else [raw_data]
        adapted_data = Adapter.funding_rate(raw_data=items)  # type: ignore[reportArgumentType]
        return adapted_data[symbol] if symbol else adapted_data

    @overload
    async def funding_interval(self, symbol: str) -> int: ...

    @overload
    async def funding_interval(self, symbol: None) -> dict[str, int]: ...

    @overload
    async def funding_interval(self) -> dict[str, int]: ...

    async def funding_interval(self, symbol: str | None = None) -> dict[str, int] | int:
        raw_data = await self._client.futures_contracts(settle="usdt")
        adapted_data = Adapter.funding_interval(raw_data=raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    @overload
    async def open_interest(self, symbol: str) -> OpenInterestItem: ...

    @overload
    async def open_interest(self, symbol: None) -> OpenInterestDict: ...

    @overload
    async def open_interest(self) -> OpenInterestDict: ...

    async def open_interest(self, symbol: str | None = None) -> OpenInterestItem | OpenInterestDict:
        raw_data = await self._client.futures_tickers(settle="usdt", contract=symbol)
        items = raw_data if isinstance(raw_data, list) else [raw_data]
        adapted_data = Adapter.open_interest(raw_data=items)  # type: ignore[reportArgumentType]
        if symbol:
            return adapted_data[symbol]
        return adapted_data

    @overload
    async def futures_best_bid_ask(self, symbol: str) -> BestBidAskItem: ...

    @overload
    async def futures_best_bid_ask(self, symbol: None) -> BestBidAskDict: ...

    @overload
    async def futures_best_bid_ask(self) -> BestBidAskDict: ...

    async def futures_best_bid_ask(
        self, symbol: str | None = None
    ) -> BestBidAskItem | BestBidAskDict:
        raw_data = await self._client.futures_tickers(settle="usdt", contract=symbol)
        adapted_data = Adapter.futures_best_bid_ask(raw_data=raw_data)  # type: ignore[reportArgumentType]
        return adapted_data[symbol] if symbol else adapted_data

    @staticmethod
    def _to_seconds(value: int | None) -> int | None:
        if value is None:
            return None
        if value >= 1_000_000_000_000:
            return value // 1000
        return value

    async def futures_depth(
        self,
        symbol: str,
        limit: int,
    ) -> BookDepthDict:
        raw_data = await self._client.futures_order_book(
            settle="usdt",
            contract=symbol,
            limit=limit,
            with_id="true",  # type: ignore
        )
        return Adapter.futures_depth(raw_data=raw_data, symbol=symbol)

    async def futures_order_create(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        quantity: str,
        price: str | None = None,
        client_order_id: str | None = None,
        reduce_only: bool | None = None,
        # Gate only args
        quantity_unit: Literal["contract", "currency"] = "contract",
    ) -> OrderIdDict:
        self.ensure_authorized()

        if type == OrderType.LIMIT and price is None:
            raise ValueError("Price is required for limit order type on Gate futures.")

        if type not in {OrderType.LIMIT, OrderType.MARKET}:
            raise ValueError(f"Unsupported order type for Gate futures: {type}.")

        if quantity_unit == "contract":
            try:
                contract_size = ExchangeInfo.get_futures_ticker_info(symbol)["contract_size"]
            except Exception as e:
                raise ValueError(f"Failed to get contract size for symbol {symbol}") from e

            quantity_decimal = Decimal(quantity)
            contract_size_decimal = Decimal(str(contract_size))
            contracts = quantity_decimal / contract_size_decimal
            if contracts % 1 != 0:
                raise ValueError(
                    f"Quantity {quantity} is not multiple of contract size {contract_size} "
                    f"for symbol {symbol}."
                )

            signed_size = contracts if side == OrderSide.BUY else -contracts
            size = str(int(signed_size))
        elif quantity_unit == "currency":
            size = quantity
        else:
            raise ValueError(f"Unsupported quantity unit: {quantity_unit}.")

        text = None
        if client_order_id:
            text = client_order_id if client_order_id.startswith("t-") else f"t-{client_order_id}"

        order_data = {
            "contract": symbol,
            "size": size,
            "price": "0" if type == OrderType.MARKET else price,
            "tif": "ioc" if type == OrderType.MARKET else "gtc",
            "text": text,
            "reduce_only": reduce_only,
        }
        raw_data = await self._client.futures_create_order(
            settle="usdt",
            order=order_data,
        )
        return Adapter.futures_order_create(raw_data)

    async def futures_position_info(self, symbol: str) -> PositionInfoDict:
        self.ensure_authorized()

        try:
            raw_data = await self._client.futures_position(
                settle="usdt",
                contract=symbol,
            )
        except ResponseError as e:
            if e.response_json.get("label") == "POSITION_NOT_FOUND":
                raw_data = {}
            else:
                raise
        return Adapter.futures_position_info(raw_data)



__all__ = ["Adapter"]

import time

from unicex.types import (
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods

from .exchange_info import ExchangeInfo


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Hyperliquid API."""

    @staticmethod
    def _resolve_spot_symbol(symbol: str, resolve_symbols: bool) -> str:
        if not resolve_symbols:
            return symbol
        try:
            return ExchangeInfo.resolve_spot_symbol(symbol)
        except ValueError:
            # Если ExchangeInfo еще не загружен, возвращаем исходное значение.
            return symbol

    @staticmethod
    def tickers(raw_data: dict, resolve_symbols: bool = True) -> list[str]:
        return [
            Adapter._resolve_spot_symbol(item["name"], resolve_symbols)
            for item in raw_data["universe"]
        ]

    @staticmethod
    def futures_tickers(raw_data: dict) -> list[str]:
        return Adapter.tickers(raw_data, resolve_symbols=False)

    @staticmethod
    def last_price(raw_data: dict, resolve_symbols: bool = True) -> dict[str, float]:
        return {
            Adapter._resolve_spot_symbol(token, resolve_symbols): float(price)
            for token, price in raw_data.items()
            if token.startswith("@")
        }

    @staticmethod
    def futures_last_price(raw_data: dict) -> dict[str, float]:
        return {
            token: float(price) for token, price in raw_data.items() if not token.startswith("@")
        }

    @staticmethod
    def ticker_24hr(raw_data: list, resolve_symbols: bool = True) -> TickerDailyDict:
        metrics = raw_data[1]
        result: TickerDailyDict = {}

        for item in metrics:
            try:
                coin = Adapter._resolve_spot_symbol(item["coin"], resolve_symbols)

                prev_day_px = float(item.get("prevDayPx") or "0")
                mid_px = float(item.get("midPx") or "0")
                mark_px = float(item.get("markPx") or "0")
                day_ntl_vlm = float(item.get("dayNtlVlm") or "0")

                p = round(((mark_px - prev_day_px) / prev_day_px * 100), 2) if prev_day_px else 0.0
                v = (day_ntl_vlm / mid_px) if mid_px else 0.0
                q = day_ntl_vlm

                if coin in result:
                    # В случае конфликта оставляем запись с большим дневным объемом.
                    prev_ticker_daily = result[coin]
                    curr_ticker_daily = TickerDailyItem(p=p, v=v, q=q)
                    if prev_ticker_daily["q"] > curr_ticker_daily["q"]:
                        result[coin] = prev_ticker_daily
                    else:
                        result[coin] = curr_ticker_daily
                else:
                    result[coin] = TickerDailyItem(p=p, v=v, q=q)

            except (KeyError, TypeError, ValueError):
                continue

        return result

    @staticmethod
    def futures_ticker_24hr(raw_data: list) -> TickerDailyDict:
        universe = raw_data[0]["universe"]
        metrics = raw_data[1]

        result: TickerDailyDict = {}

        for i, item in enumerate(metrics):
            try:
                prev_day_px = float(item.get("prevDayPx", 0) or "0")
                oracle_px = float(item.get("oraclePx", 0) or "0")
                mark_px = float(item.get("markPx", 0) or "0")
                day_ntl_vlm = float(item.get("dayNtlVlm", 0) or "0")

                p = ((mark_px - prev_day_px) / prev_day_px * 100) if prev_day_px else 0.0
                v = (day_ntl_vlm / oracle_px) if oracle_px else 0.0
                q = day_ntl_vlm

                result[universe[i]["name"]] = TickerDailyItem(p=p, v=v, q=q)
            except (KeyError, TypeError, ValueError):
                continue

        return result

    @staticmethod
    def klines(raw_data: list[dict], resolve_symbols: bool = True) -> list[KlineDict]:
        return [
            KlineDict(
                s=Adapter._resolve_spot_symbol(str(kline["s"]), resolve_symbols),
                t=kline["t"],
                o=float(kline["o"]),
                h=float(kline["h"]),
                l=float(kline["l"]),
                c=float(kline["c"]),
                v=float(kline["v"]),
                q=float(kline["v"]) * float(kline["c"]),
                T=kline["T"],
                x=None,
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x["t"]),
            )
        ]

    @staticmethod
    def futures_klines(raw_data: list[dict]) -> list[KlineDict]:
        return Adapter.klines(raw_data, resolve_symbols=False)

    @staticmethod
    def funding_rate(raw_data: list) -> dict[str, float]:
        universe = raw_data[0]["universe"]
        metrics = raw_data[1]
        return {
            universe[i]["name"]: float(item["funding"]) * 100
            for i, item in enumerate(metrics)
            if item.get("funding") is not None
        }

    @staticmethod
    def open_interest(raw_data: list) -> OpenInterestDict:
        universe = raw_data[0]["universe"]
        metrics = raw_data[1]
        return {
            universe[i]["name"]: OpenInterestItem(
                t=int(time.time() * 1000),
                v=float(item["openInterest"]),
                u="coins",
            )
            for i, item in enumerate(metrics)
        }

    @staticmethod
    def klines_message(raw_msg: dict, resolve_symbols: bool = True) -> list[KlineDict]:
        candle = raw_msg["data"]
        volume = float(candle["v"])
        close_price = float(candle["c"])

        return [
            KlineDict(
                s=Adapter._resolve_spot_symbol(str(candle["s"]), resolve_symbols),
                t=int(candle["t"]),
                o=float(candle["o"]),
                h=float(candle["h"]),
                l=float(candle["l"]),
                c=close_price,
                v=volume,
                q=volume * close_price,
                T=int(candle["T"]),
                x=None,
            )
        ]

    @staticmethod
    def trades_message(raw_msg: dict, resolve_symbols: bool = True) -> list[TradeDict]:
        result: list[TradeDict] = []
        for trade in sorted(raw_msg["data"], key=lambda item: int(item["time"])):
            side = "BUY" if trade["side"] == "B" else "SELL"
            result.append(
                TradeDict(
                    t=int(trade["time"]),
                    s=Adapter._resolve_spot_symbol(str(trade["coin"]), resolve_symbols),
                    S=side,
                    p=float(trade["px"]),
                    v=float(trade["sz"]),
                )
            )

        return result

    @staticmethod
    def best_bid_ask_message(raw_msg: dict, resolve_symbols: bool = True) -> list[BestBidAskItem]:
        data = raw_msg["data"]
        best_bid = data["bbo"][0]
        best_ask = data["bbo"][1]

        bid_price = float(best_bid["px"]) if best_bid else 0.0
        bid_size = float(best_bid["sz"]) if best_bid else 0.0
        ask_price = float(best_ask["px"]) if best_ask else 0.0
        ask_size = float(best_ask["sz"]) if best_ask else 0.0

        return [
            BestBidAskItem(
                s=Adapter._resolve_spot_symbol(str(data["coin"]), resolve_symbols),
                t=int(data["time"]),
                u=int(data["time"]),
                b=bid_price,
                B=bid_size,
                a=ask_price,
                A=ask_size,
            )
        ]

    @staticmethod
    def partial_book_depth_message(
        raw_msg: dict,
        limit: int,
        resolve_symbols: bool = True,
    ) -> list[BookDepthDict]:
        data = raw_msg["data"]
        bids = data["levels"][0][:limit]
        asks = data["levels"][1][:limit]

        return [
            BookDepthDict(
                s=Adapter._resolve_spot_symbol(str(data["coin"]), resolve_symbols),
                t=int(data["time"]),
                u=int(data["time"]),
                b=[(float(level["px"]), float(level["sz"])) for level in bids],
                a=[(float(level["px"]), float(level["sz"])) for level in asks],
            )
        ]



__all__ = ["UniWebsocketManager"]

from collections.abc import Awaitable, Callable, Sequence
from typing import Any

from unicex._abc import IUniWebsocketManager
from unicex._base import Websocket
from unicex.enums import Exchange, Timeframe
from unicex.exceptions import NotSupported
from unicex.types import LoggerLike
from unicex.utils import validate_allowed_kwargs

from .adapter import Adapter
from .client import Client
from .uni_client import UniClient
from .websocket_manager import WebsocketManager

type CallbackType = Callable[[Any], Awaitable[None]]


class UniWebsocketManager(IUniWebsocketManager):
    """Реализация менеджера асинхронных унифицированных вебсокетов."""

    def __init__(
        self,
        client: Client | UniClient | None = None,
        logger: LoggerLike | None = None,
        **ws_kwargs: Any,
    ) -> None:
        """Инициализирует унифицированный менеджер вебсокетов.

        Параметры:
            client (`Client | UniClient | None`): Клиент Aster или унифицированный клиент. Нужен для подключения к приватным топикам.
            logger (`LoggerLike | None`): Логгер для записи логов.
            ws_kwargs (`dict[str, Any]`): Дополнительные параметры инициализации, которые будут переданы WebsocketManager/Websocket.
        """
        super().__init__(client=client, logger=logger)
        self._websocket_manager = WebsocketManager(self._client, **ws_kwargs)  # type: ignore
        self._adapter = Adapter()

    def klines(
        self,
        callback: CallbackType,
        timeframe: Timeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        raise NotSupported("Spot market data is not supported for Aster")

    def futures_klines(
        self,
        callback: CallbackType,
        timeframe: Timeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        wrapper = self._make_wrapper(self._adapter.Klines_message, callback)
        return self._websocket_manager.futures_klines(
            callback=wrapper,
            symbol=symbol,
            symbols=symbols,
            interval=timeframe.to_exchange_format(Exchange.ASTER),
        )

    def trades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        raise NotSupported("Spot market data is not supported for Aster")

    def aggtrades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        raise NotSupported("Spot market data is not supported for Aster")

    def futures_trades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        return self.futures_aggtrades(callback, symbol, symbols)  # type: ignore

    def futures_aggtrades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        wrapper = self._make_wrapper(self._adapter.trades_message, callback)
        return self._websocket_manager.futures_agg_trade(
            callback=wrapper, symbol=symbol, symbols=symbols
        )

    def futures_best_bid_ask(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        wrapper = self._make_wrapper(self._adapter.futures_best_bid_ask_message, callback)
        return self._websocket_manager.futures_symbol_book_ticker(
            callback=wrapper,
            symbol=symbol,
            symbols=symbols,
        )

    def futures_partial_book_depth(
        self,
        callback: CallbackType,
        limit: int,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
        **kwargs: Any,
    ) -> Websocket:
        allowed_levels = {5, 10, 20}
        if limit not in allowed_levels:
            raise ValueError("Parameter `limit` must be one of: 5, 10, 20")

        validate_allowed_kwargs(kwargs=kwargs, allowed_kwargs=("update_speed",))

        update_speed: str | None = kwargs.get("update_speed")
        allowed_update_speeds = {None, "100ms", "500ms"}
        if update_speed not in allowed_update_speeds:
            raise ValueError("Parameter `update_speed` must be one of: '100ms', '500ms' or None")

        wrapper = self._make_wrapper(self._adapter.futures_partial_book_depth_message, callback)
        return self._websocket_manager.futures_partial_book_depth(
            callback=wrapper,
            symbol=symbol,
            symbols=symbols,
            levels=str(limit),
            update_speed=update_speed,
        )



__all__ = ["UniClient"]


from typing import overload

from unicex._abc import IUniClient
from unicex.enums import Exchange, OrderSide, OrderType, Timeframe
from unicex.types import (
    BestBidAskDict,
    BestBidAskItem,
    BookDepthDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    OrderIdDict,
    PositionInfoDict,
    TickerDailyDict,
)

from .adapter import Adapter
from .client import Client


class UniClient(IUniClient[Client]):
    """Унифицированный клиент для работы с Kucoin API."""

    @property
    def _client_cls(self) -> type[Client]:
        return Client

    async def tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.ticker("SPOT")
        return Adapter.tickers(raw_data, only_usdt)

    async def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        raw_data = await self._client.ticker("FUTURES")
        return Adapter.futures_tickers(raw_data, only_usdt)

    async def last_price(self) -> dict[str, float]:
        raw_data = await self._client.ticker("SPOT")
        return Adapter.last_price(raw_data)

    async def futures_last_price(self) -> dict[str, float]:
        raw_data = await self._client.ticker("FUTURES")
        return Adapter.last_price(raw_data)

    async def ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.ticker("SPOT")
        return Adapter.ticker_24hr(raw_data)

    async def futures_ticker_24hr(self) -> TickerDailyDict:
        raw_data = await self._client.ticker("FUTURES")
        return Adapter.ticker_24hr(raw_data)

    async def klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        if not limit and not all([start_time, end_time]):
            raise ValueError("limit or (start_time and end_time) must be provided")

        if limit:  # Перезаписываем start_time и end_time если указан limit, т.к. по умолчанию HyperLiquid не принимают этот параметр
            if not isinstance(interval, Timeframe):
                raise ValueError("interval must be a Timeframe if limit param provided")
            start_time, end_time = self.limit_to_start_and_end_time(
                interval, limit, use_milliseconds=False
            )
        interval = (
            interval.to_exchange_format(Exchange.KUCOIN)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.kline(
            trade_type="SPOT",
            symbol=symbol,
            interval=interval,
            start_at=self.to_seconds(start_time),
            end_at=self.to_seconds(end_time),
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)

    async def futures_klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        if not limit and not all([start_time, end_time]):
            raise ValueError("limit or (start_time and end_time) must be provided")

        if limit:  # Перезаписываем start_time и end_time если указан limit, т.к. по умолчанию HyperLiquid не принимают этот параметр
            if not isinstance(interval, Timeframe):
                raise ValueError("interval must be a Timeframe if limit param provided")
            start_time, end_time = self.limit_to_start_and_end_time(
                interval, limit, use_milliseconds=False
            )
        interval = (
            interval.to_exchange_format(Exchange.KUCOIN)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.kline(
            trade_type="FUTURES",
            symbol=symbol,
            interval=interval,
            start_at=self.to_seconds(start_time),
            end_at=self.to_seconds(end_time),
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)

    async def funding_rate(self, symbol: str | None = None) -> dict[str, float] | float:
        if not symbol:
            raise ValueError("Symbol is required to fetch Kucoin funding rate")
        raw_data = await self._client.funding_rate(symbol=symbol)
        return Adapter.funding_rate(raw_data)

    @overload
    async def funding_interval(self, symbol: str) -> int: ...

    @overload
    async def funding_interval(self, symbol: None) -> dict[str, int]: ...

    @overload
    async def funding_interval(self) -> dict[str, int]: ...

    async def funding_interval(self, symbol: str | None = None) -> dict[str, int] | int:
        raise NotImplementedError("This method will be implemented in a future release")

    @overload
    async def open_interest(self, symbol: str) -> OpenInterestItem: ...

    @overload
    async def open_interest(self, symbol: None) -> OpenInterestDict: ...

    @overload
    async def open_interest(self) -> OpenInterestDict: ...

    async def open_interest(self, symbol: str | None = None) -> OpenInterestItem | OpenInterestDict:
        raw_data = await self._client.open_interest()
        adapted_data = Adapter.open_interest(raw_data)
        return adapted_data[symbol] if symbol else adapted_data

    async def futures_best_bid_ask(
        self, symbol: str | None = None
    ) -> BestBidAskItem | BestBidAskDict:
        raise NotImplementedError("Method `futures_best_bid_ask` will be implemented later")

    async def futures_depth(
        self,
        symbol: str,
        limit: int,
    ) -> BookDepthDict:
        raise NotImplementedError(
            "Method 'futures_depth' will be implemented later. You can open pull request to contribute."
        )

    async def futures_order_create(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        quantity: str,
        price: str | None = None,
        client_order_id: str | None = None,
        reduce_only: bool | None = None,
    ) -> OrderIdDict:
        raise NotImplementedError("Method will be implemented later.")

    async def futures_position_info(self, symbol: str) -> PositionInfoDict:
        raise NotImplementedError("Method will be implemented later.")



__all__ = ["IUniWebsocketManager"]

from collections.abc import Awaitable, Callable, Sequence
from typing import Any

from unicex._abc import IUniWebsocketManager
from unicex._base import Websocket
from unicex.enums import Exchange, Timeframe
from unicex.types import LoggerLike

from .adapter import Adapter
from .client import Client
from .uni_client import UniClient
from .websocket_manager import WebsocketManager

type CallbackType = Callable[[Any], Awaitable[None]]


class UniWebsocketManager(IUniWebsocketManager):
    """Реализация менеджера асинхронных унифицированных вебсокетов."""

    def __init__(
        self,
        client: Client | UniClient | None = None,
        logger: LoggerLike | None = None,
        **ws_kwargs: Any,
    ) -> None:
        """Инициализирует унифицированный менеджер вебсокетов.

        Параметры:
            client (`Client | UniClient | None`): Клиент Hyperliquid или унифицированный клиент. Нужен для подключения к приватным топикам.
            logger (`LoggerLike | None`): Логгер для записи логов.
            ws_kwargs (`dict[str, Any]`): Дополнительные параметры инициализации, которые будут переданы WebsocketManager/Websocket.
        """
        super().__init__(client=client, logger=logger)
        self._websocket_manager = WebsocketManager(self._client, **ws_kwargs)  # type: ignore
        self._adapter = Adapter()

    def _is_service_message(self, raw_msg: Any) -> bool:
        return raw_msg.get("channel") in {"subscriptionResponse", "pong"}

    def klines(
        self,
        callback: CallbackType,
        timeframe: Timeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
        resolve_symbols: bool = True,
    ) -> Websocket:
        wrapper = self._make_wrapper(
            lambda raw_msg: self._adapter.klines_message(
                raw_msg=raw_msg,
                resolve_symbols=resolve_symbols,
            ),
            callback,
        )
        return self._websocket_manager.candle(
            callback=wrapper,
            interval=timeframe.to_exchange_format(Exchange.HYPERLIQUID),  # type: ignore
            coin=symbol,
            coins=list(symbols) if symbols else None,
        )

    def futures_klines(
        self,
        callback: CallbackType,
        timeframe: Timeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        return self.klines(
            callback=callback,
            timeframe=timeframe,
            symbol=symbol,  # type: ignore
            symbols=symbols,  # type: ignore
            resolve_symbols=False,
        )

    def trades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
        resolve_symbols: bool = True,
    ) -> Websocket:
        wrapper = self._make_wrapper(
            lambda raw_msg: self._adapter.trades_message(
                raw_msg=raw_msg,
                resolve_symbols=resolve_symbols,
            ),
            callback,
        )
        return self._websocket_manager.trades(
            callback=wrapper,
            coin=symbol,
            coins=list(symbols) if symbols else None,
        )

    def aggtrades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
        resolve_symbols: bool = True,
    ) -> Websocket:
        return self.trades(
            callback=callback,
            symbol=symbol,  # type: ignore
            symbols=symbols,  # type: ignore
            resolve_symbols=resolve_symbols,
        )

    def futures_trades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        return self.trades(
            callback=callback,
            symbol=symbol,  # type: ignore
            symbols=symbols,  # type: ignore
            resolve_symbols=False,
        )

    def futures_aggtrades(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        return self.futures_trades(callback=callback, symbol=symbol, symbols=symbols)  # type: ignore

    def futures_best_bid_ask(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        wrapper = self._make_wrapper(
            lambda raw_msg: self._adapter.best_bid_ask_message(
                raw_msg=raw_msg,
                resolve_symbols=False,
            ),
            callback,
        )
        return self._websocket_manager.bbo(
            callback=wrapper,
            coin=symbol,
            coins=list(symbols) if symbols else None,
        )

    def futures_partial_book_depth(
        self,
        callback: CallbackType,
        limit: int,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        if limit <= 0:
            raise ValueError("Parameter `limit` must be greater than 0")
        if limit > 20:
            raise ValueError("Parameter `limit` must be less than or equal to 20")

        wrapper = self._make_wrapper(
            lambda raw_msg: self._adapter.partial_book_depth_message(
                raw_msg=raw_msg,
                limit=limit,
                resolve_symbols=False,
            ),
            callback,
        )
        return self._websocket_manager.l2_book(
            callback=wrapper,
            coin=symbol,
            coins=list(symbols) if symbols else None,
        )
